################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import json
import random
from perspective import Table, PerspectiveManager

data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}


class TestPerspectiveSession(object):
    def post(self, msg):
        '''boilerplate callback to simulate a client's `post()` method.'''
        msg = json.loads(msg)
        assert msg["id"] is not None

    def validate_post(self, msg, expected=None):
        msg = json.loads(msg)
        if expected:
            assert msg == expected

    # test session
    def test_session_new_session(self, sentinel):
        s = sentinel(False)

        def handle_to_dict(msg):
            s.set(True)
            message = json.loads(msg)
            assert message["data"] == data

        message = {"id": 1, "table_name": "table1", "view_name": "view1", "cmd": "view"}

        manager = PerspectiveManager()
        session = manager.new_session()
        client_id = session.client_id
        table = Table(data)

        manager.host_table("table1", table)

        # create a view through the session to make sure it has a client id
        session.process(message, self.post)

        # make sure the client ID is attached to the new view
        assert len(manager._views.keys()) == 1
        assert manager._get_view("view1")._client_id == client_id

        to_dict_message = {"id": 2, "name": "view1", "cmd": "view_method", "method": "to_dict"}

        session.process(to_dict_message, handle_to_dict)
        assert s.get() is True

    def test_session_multiple_new_sessions(self, sentinel):
        s = sentinel(0)

        def handle_to_dict(msg):
            s.set(s.get() + 1)

            message = json.loads(msg)

            assert message["data"] == {
                "a": [1, 2, 3, 1, 2, 3],
                "b": ["a", "b", "c", "str1", "str2", "str3"]
            }

        manager = PerspectiveManager()
        sessions = [manager.new_session() for i in range(5)]
        table = Table(data)

        manager.host_table("table1", table)

        # create a view on each session
        for i, session in enumerate(sessions):
            # IDs have to conflict - each viewer will send the first message as
            # ID = 1, so we need to make sure we handle that.
            msg = {"id": 1, "table_name": "table1", "view_name": "view" + str(i), "cmd": "view"}
            session.process(msg, self.post)

        manager_views = list(manager._views.keys())
        for key in ["view" + str(i) for i in range(5)]:
            assert key in manager_views

        for i, session in enumerate(sessions):
            view = manager._get_view("view" + str(i))
            assert view._client_id == session.client_id

        # arbitrarily do an update
        random_session_id = random.randint(0, 4)
        update_message = {"id": 2, "name": "table1", "cmd": "table_method", "method": "update", "args": [{"a": [1, 2, 3], "b": ["str1", "str2", "str3"]}]}
        sessions[random_session_id].process(update_message, self.post)

        # should reflect in all sessions
        for i, session in enumerate(sessions):
            to_dict_message = {"id": 3, "name": "view" + str(i), "cmd": "view_method", "method": "to_dict"}
            session.process(to_dict_message, handle_to_dict)

        assert s.get() == 5

    def test_session_close_session_with_callbacks(self, sentinel):
        s = sentinel(0)

        manager = PerspectiveManager()
        session = manager.new_session()
        client_id = session.client_id

        # create a table and view using manager
        make_table = {"id": 1, "name": "table1", "cmd": "table", "args": [data]}
        session.process(make_table, self.post)

        make_view = {"id": 2, "table_name": "table1", "view_name": "view1", "cmd": "view"}
        session.process(make_view, self.post)

        # make sure the client ID is attached to the new view
        assert len(manager._views.keys()) == 1
        assert manager._get_view("view1")._client_id == client_id

        def callback(updated):
            assert updated["port_id"] == 0
            s.set(s.get() + 100)

        # simulate a client that holds callbacks by id
        callbacks = {
            3: callback
        }

        def post_update(msg):
            # when `on_update` is triggered, this callback gets the message
            # and has to decide which callback to trigger.
            message = json.loads(msg)
            assert message["id"] is not None
            if message["id"] == 3:
                # trigger callback
                assert message["data"] == {
                    "port_id": 0
                }
                callbacks[message["id"]](message["data"])

        # hook into the created view and pass it the callback
        make_on_update = {"id": 3, "name": "view1", "cmd": "view_method", "subscribe": True, "method": "on_update", "callback_id": "callback_1"}
        session.process(make_on_update, post_update)

        # call updates
        update1 = {"id": 4, "name": "table1", "cmd": "table_method", "method": "update", "args": [{"a": [4], "b": ["d"]}]}
        update2 = {"id": 5, "name": "table1", "cmd": "table_method", "method": "update", "args": [{"a": [5], "b": ["e"]}]}

        session.process(update1, self.post)
        session.process(update2, self.post)

        assert s.get() == 200

        # close the session
        session.close()

        # make sure the view is gone - but not the table
        assert "table1" in manager._tables
        assert manager._views == {}
        assert len(manager._callback_cache) == 0

    def test_session_close_multiple_sessions_with_callbacks(self, sentinel):
        s = sentinel(0)

        manager = PerspectiveManager()
        sessions = [manager.new_session() for i in range(5)]

        # create a table and view using manager
        make_table = {"id": 1, "name": "table1", "cmd": "table", "args": [data]}
        manager._process(make_table, self.post)

        # create a view on each session
        for i, session in enumerate(sessions):
            # IDs have to conflict - each viewer will send the first message as
            # ID = 1, so we need to make sure we handle that.
            msg = {"id": 2, "table_name": "table1", "view_name": "view" + str(i), "cmd": "view"}
            session.process(msg, self.post)

        manager_views = list(manager._views.keys())
        for key in ["view" + str(i) for i in range(5)]:
            assert key in manager_views

        for i, session in enumerate(sessions):
            view = manager._get_view("view" + str(i))
            assert view._client_id == session.client_id

        def callback(updated):
            assert updated["port_id"] == 0
            s.set(s.get() + 100)

        # simulate a client that holds callbacks by id
        callbacks = {
            3: callback
        }

        def post_update(msg):
            # when `on_update` is triggered, this callback gets the message
            # and has to decide which callback to trigger.
            message = json.loads(msg)
            assert message["id"] is not None
            if message["id"] == 3:
                # trigger callback
                assert message["data"] == {
                    "port_id": 0
                }
                callbacks[message["id"]](message["data"])

        # create a view and an on_update on each session
        for i, session in enumerate(sessions):
            view_name = "view" + str(i)
            # IDs have to conflict - each viewer will send the first message as
            # ID = 1, so we need to make sure we handle that.
            msg = {"id": 2, "table_name": "table1", "view_name": view_name, "cmd": "view"}
            session.process(msg, self.post)
            make_on_update = {"id": 3, "name": view_name, "cmd": "view_method", "subscribe": True, "method": "on_update", "callback_id": "callback_1"}
            session.process(make_on_update, post_update)

        # call updates using a random session - they should propagate
        random_session_id = random.randint(0, 4)
        random_session = sessions[random_session_id]
        random_client_id = random_session.client_id

        update1 = {"id": 4, "name": "table1", "cmd": "table_method", "method": "update", "args": [{"a": [4], "b": ["d"]}]}
        update2 = {"id": 5, "name": "table1", "cmd": "table_method", "method": "update", "args": [{"a": [5], "b": ["e"]}]}

        random_session.process(update1, self.post)
        random_session.process(update2, self.post)

        # all updates processed, all callbacks fired
        assert s.get() == 1000

        # close a random session, and make sure the other views and callbacks
        # are not affected
        random_session.close()

        # make sure the view is gone - but not the table
        assert "table1" in manager._tables

        assert "view" + str(random_session_id) not in manager._views.keys()
        assert len(manager._views.keys()) == 4
        
        for callback in manager._callback_cache:
            assert callback["client_id"] != random_client_id

        assert len(manager._callback_cache) == 4


################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import asyncio
from tornado.websocket import WebSocketHandler

from .common import PerspectiveHandlerBase


class PerspectiveTornadoHandler(PerspectiveHandlerBase, WebSocketHandler):
    """PerspectiveTornadoHandler is a drop-in implementation of Perspective.

    Use it inside Tornado routing to create a server-side Perspective that is
    ready to receive websocket messages from the front-end `perspective-viewer`.
    Because Tornado implements an event loop, this handler links Perspective
    with `IOLoop.current()` in order to defer expensive operations until the
    next free iteration of the event loop.

    The Perspective client and server will automatically keep the Websocket
    alive without timing out.

    Examples:
        >>> MANAGER = PerspectiveManager()
        >>> MANAGER.host_table("data_source_one", Table(
        ...     pd.read_csv("superstore.csv")))
        >>> app = tornado.web.Application([
        ...     (r"/", MainHandler),
        ...     (r"/websocket", PerspectiveTornadoHandler, {
        ...         "manager": MANAGER,
        ...         "check_origin": True
        ...     })
        ... ])
    """

    def __init__(self, *args, **kwargs):
        PerspectiveHandlerBase.__init__(self, **kwargs)
        WebSocketHandler.__init__(self, *args)

    def on_message(self, *args, **kwargs):
        return asyncio.ensure_future(
            PerspectiveHandlerBase.on_message(self, *args, **kwargs)
        )

    async def write_message(self, message: str, binary: bool = False) -> None:
        return WebSocketHandler.write_message(self, message=message, binary=binary)

    # Use common docstring
    __init__.__doc__ = PerspectiveHandlerBase.__init__.__doc__



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

# NOTE: This is essentially a clone of the tornado handler tests,
# but using async/await and starlette handler
import random
import pytest

from aiohttp import web
from datetime import datetime

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveAIOHTTPHandler,
    aiohttp_websocket as websocket,
)


data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()


async def websocket_handler(request):
    handler = PerspectiveAIOHTTPHandler(
        manager=MANAGER, request=request, chunk_size=500
    )
    await handler.run()


@pytest.fixture
def app():
    app = web.Application()
    app.router.add_get("/websocket", websocket_handler)
    return app


class TestPerspectiveAIOHTTPHandlerChunked(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self, app, aiohttp_client):
        """Connect and initialize a websocket client connection to the
        Perspective aiottp server.
        """
        client = await aiohttp_client(app)
        return await websocket(
            "http://{}:{}/websocket".format(client.host, client.port), client.session
        )

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_to_arrow_chunked(
        self, app, aiohttp_client
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()
        output = await view.to_arrow()
        expected = await table.schema()

        assert Table(output).schema(as_string=True) == expected

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_to_arrow_update_chunked(
        self, app, aiohttp_client
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        output = await view.to_arrow()

        for i in range(10):
            await table.update(output)

        size2 = await table.size()
        assert size2 == 110

    @pytest.mark.asyncio
    async def test_aiohttp_handler_update_chunked_interleaved_with_trivial(
        self, app, aiohttp_client
    ):
        """Tests that, when a chunked response `output_fut` is interleaved with
        a response belonging to another message ID (and not binary encoded)
        `size3`, both messages de-multiplex correclty and succeed.
        """
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        output_fut = view.to_arrow()
        size3 = await view.num_rows()
        assert size3 == 10

        output = await output_fut

        for i in range(10):
            await table.update(output)

        size2 = await table.size()
        assert size2 == 110



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import os
from datetime import date, datetime
from io import StringIO

import numpy as np
import pandas as pd
import pytz
from perspective.table import Table
from pytest import mark

IS_WIN = os.name == 'nt'


class TestToFormat(object):

    # to_records

    def test_to_records_int(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == data

    def test_to_records_float(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == data

    def test_to_records_string(self):
        data = [{"a": "string1", "b": "string2"}, {"a": "string3", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == data

    def test_to_records_date(self):
        today = date.today()
        dt = datetime(today.year, today.month, today.day)
        data = [{"a": today, "b": "string2"}, {"a": today, "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == [{"a": dt, "b": "string2"}, {"a": dt, "b": "string4"}]

    def test_to_records_date_no_dst(self):
        # make sure that DST does not affect the way we read dates - if tm_dst in `t_date::get_tm()` isn't set to -1, it could reverse 1hr by assuming DST is not in effect.
        today = date.today()
        dt = datetime(today.year, today.month, today.day)
        data = [{"a": today, "b": "string2"}, {"a": today, "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == [{"a": dt, "b": "string2"}, {"a": dt, "b": "string4"}] 

    def test_to_records_date_str(self):
        data = [{"a": "03/11/2019", "b": "string2"}, {"a": "03/12/2019", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == [{"a": datetime(2019, 3, 11), "b": "string2"}, {"a": datetime(2019, 3, 12), "b": "string4"}]

    def test_to_records_date_str_month_first(self):
        data = [{"a": "1/2/2019", "b": "string2"}, {"a": "3/4/2019", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.schema() == {"a": date, "b": str}
        assert view.to_records() == [{"a": datetime(2019, 1, 2), "b": "string2"}, {"a": datetime(2019, 3, 4), "b": "string4"}]

    def test_to_records_date_str_month_ymd(self):
        data = [{"a": "2019/01/02", "b": "string2"}, {"a": "2019/03/04", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.schema() == {"a": date, "b": str}
        assert view.to_records() == [{"a": datetime(2019, 1, 2), "b": "string2"}, {"a": datetime(2019, 3, 4), "b": "string4"}]

    def test_to_records_datetime(self):
        dt = datetime(2019, 9, 10, 19, 30, 59, 515000)
        data = [{"a": dt, "b": "string2"}, {"a": dt, "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == data  # should have symmetric input/output

    def test_to_records_datetime_str(self):
        data = [{"a": "03/11/2019 3:15PM", "b": "string2"}, {"a": "3/11/2019 3:20PM", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == [{"a": datetime(2019, 3, 11, 15, 15), "b": "string2"}, {"a": datetime(2019, 3, 11, 15, 20), "b": "string4"}]

    def test_to_records_datetime_str_tz(self):
        dt = "2019/07/25T15:30:00+00:00"
        data = [{"a": dt}, {"a": dt}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records()
        for r in records:
            r["a"] = r["a"].replace(tzinfo=pytz.utc)
        assert records == [{"a": datetime(2019, 7, 25, 15, 30, tzinfo=pytz.utc)}, {"a": datetime(2019, 7, 25, 15, 30, tzinfo=pytz.utc)}]

    def test_to_records_datetime_ms_str(self):
        data = [{"a": "03/11/2019 3:15:15.999PM"}, {"a": "3/11/2019 3:15:16.001PM"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == [{"a": datetime(2019, 3, 11, 15, 15, 15, 999000)}, {"a": datetime(2019, 3, 11, 15, 15, 16, 1000)}]

    def test_to_records_none(self):
        data = [{"a": None, "b": 1}, {"a": None, "b": 2}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records() == data

    def test_to_records_one(self):
        data = [{"a": 1, "b": "string1"}, {"a": 1, "b": "string2"}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        assert view.to_records() == [
            {"__ROW_PATH__": [], "a": 2, "b": 2}, {"__ROW_PATH__": [1], "a": 2, "b": 2}
        ]

    def test_to_records_two(self):
        data = [{"a": 1, "b": "string1"}, {"a": 1, "b": "string2"}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        assert view.to_records() == [
            {"__ROW_PATH__": [], "string1|a": 1, "string1|b": 1, "string2|a": 1, "string2|b": 1},
            {"__ROW_PATH__": [1], "string1|a": 1, "string1|b": 1, "string2|a": 1, "string2|b": 1},
        ]

    def test_to_records_column_only(self):
        data = [{"a": 1, "b": "string1"}, {"a": 1, "b": "string2"}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"]
        )
        assert view.to_records() == [
            {"string1|a": 1, "string1|b": "string1", "string2|a": None, "string2|b": None},
            {"string1|a": None, "string1|b": None, "string2|a": 1, "string2|b": "string2"},
        ]

    # to_dict

    def test_to_dict_int(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [1, 3],
            "b": [2, 4]
        }

    def test_to_dict_float(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [1.5, 3.5],
            "b": [2.5, 4.5]
        }

    def test_to_dict_date(self):
        today = date.today()
        dt = datetime(today.year, today.month, today.day)
        data = [{"a": today, "b": 2}, {"a": today, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [dt, dt],
            "b": [2, 4]
        }

    def test_to_dict_datetime(self):
        dt = datetime(2019, 3, 15, 20, 30, 59, 6000)
        data = [{"a": dt, "b": 2}, {"a": dt, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [dt, dt],
            "b": [2, 4]
        }

    def test_to_dict_bool(self):
        data = [{"a": True, "b": False}, {"a": True, "b": False}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [True, True],
            "b": [False, False]
        }

    def test_to_dict_string(self):
        data = [{"a": "string1", "b": "string2"}, {"a": "string3", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": ["string1", "string3"],
            "b": ["string2", "string4"]
        }

    def test_to_dict_none(self):
        data = [{"a": None, "b": None}, {"a": None, "b": None}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict() == {
            "a": [None, None],
            "b": [None, None]
        }

    def test_to_dict_one(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        assert view.to_dict() == {
            "__ROW_PATH__": [[], [1]],
            "a": [2, 2],
            "b": [4, 4]
        }

    def test_to_dict_two(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        assert view.to_dict() == {
            "__ROW_PATH__": [[], [1]],
            "2|a": [2, 2],
            "2|b": [4, 4]
        }

    def test_to_dict_column_only(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"]
        )
        assert view.to_dict() == {
            "2|a": [1, 1],
            "2|b": [2, 2],
        }

    def test_to_dict_one_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            columns=[]
        )
        assert view.to_dict() == {"__ROW_PATH__": [[], [1]]}

    def test_to_dict_two_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            columns=[]
        )
        assert view.to_dict() == {
            "__ROW_PATH__": [[], [1]]
        }

    def test_to_dict_column_only_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"],
            columns=[]
        )
        assert view.to_dict() == {}

    # to_numpy

    def test_to_numpy_int(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([1, 3]))
        assert np.array_equal(v["b"], np.array([2, 4]))

    def test_to_numpy_float(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([1.5, 3.5]))
        assert np.array_equal(v["b"], np.array([2.5, 4.5]))

    def test_to_numpy_bool(self):
        data = [{"a": True, "b": False}, {"a": True, "b": False}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([True, True]))
        assert np.array_equal(v["b"], np.array([False, False]))

    def test_to_numpy_date(self):
        today = date.today()
        dt = datetime(today.year, today.month, today.day)
        data = [{"a": dt, "b": 2}, {"a": dt, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([dt, dt]))

    def test_to_numpy_datetime(self):
        dt = datetime(2019, 3, 15, 20, 30, 59, 6000)
        data = [{"a": dt}, {"a": dt}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([dt, dt]))

    def test_to_numpy_string(self):
        data = [{"a": "string1", "b": "string2"}, {"a": "string3", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array(["string1", "string3"]))
        assert np.array_equal(v["b"], np.array(["string2", "string4"]))

    def test_to_numpy_none(self):
        data = [{"a": None, "b": None}, {"a": None, "b": None}]
        tbl = Table(data)
        view = tbl.view()
        v = view.to_numpy()
        assert np.array_equal(v["a"], np.array([None, None]))
        assert np.array_equal(v["b"], np.array([None, None]))

    def test_to_numpy_one(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        v = view.to_numpy()
        assert np.array_equal(v["__ROW_PATH__"], [[], [1]])
        assert np.array_equal(v["a"], np.array([2, 2]))
        assert np.array_equal(v["b"], np.array([4, 4]))

    def test_to_numpy_two(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        v = view.to_numpy()
        assert np.array_equal(v["__ROW_PATH__"], [[], [1]])
        assert np.array_equal(v["2|a"], np.array([2, 2]))
        assert np.array_equal(v["2|b"], np.array([4, 4]))

    def test_to_numpy_column_only(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"]
        )
        v = view.to_numpy()
        assert np.array_equal(v["2|a"], np.array([1, 1]))
        assert np.array_equal(v["2|b"], np.array([2, 2]))

    def test_to_pandas_df_simple(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        df = pd.DataFrame(data)
        tbl = Table(df)
        view = tbl.view()
        df2 = view.to_df()
        assert np.array_equal(df2.columns, pd.Index(["index", "a", "b"], dtype=object))
        assert np.array_equal(df2["a"].values, df["a"].values)
        assert np.array_equal(df2["b"].values, df["b"].values)

    def test_to_pandas_df_simple_series(self):
        inp = pd.Series([1, 2, 3], name="a")
        df = pd.DataFrame()
        df["a"] = pd.Series([1, 2, 3])
        tbl = Table(inp)
        view = tbl.view()
        df2 = view.to_df()
        assert np.array_equal(df2.columns, pd.Index(["index", "a"], dtype=object))
        assert np.array_equal(df2["a"].values, df["a"].values)

    # start_row/end_row
    def test_to_records_zero_over_max_row(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_row=1000
        )
        assert records == data

    def test_to_records_one_over_max_row(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        records = view.to_records(
            end_row=1000
        )
        assert records == [
            {'__ROW_PATH__': [], 'a': 5, 'b': 7},
            {'__ROW_PATH__': [1.5], 'a': 1.5, 'b': 2.5},
            {'__ROW_PATH__': [3.5], 'a': 3.5, 'b': 4.5}
        ]

    def test_to_records_two_over_max_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=1000
        )
        assert records == [
            {'2|a': 1, '2|b': 2, '4|a': 3, '4|b': 4, '__ROW_PATH__': []},
            {'2|a': 1, '2|b': 2, '4|a': None, '4|b': None, '__ROW_PATH__': [1]},
            {'2|a': None, '2|b': None, '4|a': 3, '4|b': 4, '__ROW_PATH__': [3]}
        ]

    def test_to_records_start_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1
        )
        assert records == [{"a": 3, "b": 4}]

    def test_to_records_end_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_row=1
        )
        assert records == [{"a": 1, "b": 2}]

    def test_to_records_start_row_end_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1,
            end_row=2
        )
        assert records == [{"a": 3, "b": 4}]

    def test_to_records_start_row_end_row_equiv(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1,
            end_row=1
        )
        assert records == []

    def test_to_records_floor_start_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1.5
        )
        assert records == [{"a": 3, "b": 4}]

    def test_to_records_ceil_end_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_row=0.5
        )
        assert records == [{"a": 1, "b": 2}]

    def test_to_records_floor_start_row_ceil_end_row(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1.5,
            end_row=1.5
        )
        assert records == [{"a": 3, "b": 4}]

    def test_to_records_floor_start_row_ceil_end_row_equiv(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_row=1.5,
            end_row=0.5
        )
        assert records == []

    # start_col/end_col

    def test_to_records_zero_over_max_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_col=1000
        )
        assert records == data

    def test_to_records_zero_start_gt_end_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=2,
            end_col=1
        )
        assert records == [{}, {}]

    def test_to_records_zero_start_eq_end_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1,
            end_col=1
        )
        assert records == [{}, {}]

    def test_to_records_one_over_max_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        records = view.to_records(
            end_col=1000
        )
        assert records == [
            {'__ROW_PATH__': [], 'a': 5, 'b': 7},
            {'__ROW_PATH__': [1.5], 'a': 1.5, 'b': 2.5},
            {'__ROW_PATH__': [3.5], 'a': 3.5, 'b': 4.5}
        ]

    def test_to_records_one_start_gt_end_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        records = view.to_records(
            start_col=2,
            end_col=1
        )
        assert records == [{}, {}, {}]

    def test_to_records_one_start_gt_end_col_large(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        records = view.to_records(
            start_col=20,
            end_col=19
        )
        assert records == [{}, {}, {}]

    def test_to_records_one_start_eq_end_col(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        records = view.to_records(
            start_col=0,
            end_col=0
        )
        assert records == [{'__ROW_PATH__': []}, {'__ROW_PATH__': [1.5]}, {'__ROW_PATH__': [3.5]}]

    def test_to_records_two_over_max_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_col=1000
        )
        assert records == [
            {'2|a': 1, '2|b': 2, '4|a': 3, '4|b': 4, '__ROW_PATH__': []},
            {'2|a': 1, '2|b': 2, '4|a': None, '4|b': None, '__ROW_PATH__': [1]},
            {'2|a': None, '2|b': None, '4|a': 3, '4|b': 4, '__ROW_PATH__': [3]}
        ]

    def test_to_records_start_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1
        )
        assert records == [{"b": 2}, {"b": 4}]

    def test_to_records_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_col=1
        )
        assert records == [{"a": 1}, {"a": 3}]

    def test_to_records_two_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=12,
            end_col=5
        )
        assert records == [
            {'2|a': 1, '2|b': 2, '4|a': 3, '4|b': 4, '__ROW_PATH__': []},
            {'2|a': 1, '2|b': 2, '4|a': None, '4|b': None, '__ROW_PATH__': [1]},
            {'2|a': None, '2|b': None, '4|a': 3, '4|b': 4, '__ROW_PATH__': [3]}
        ]

    def test_to_records_two_start_gt_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=12,
            start_col=5,
            end_col=4
        )
        assert records == [{}, {}, {}]

    def test_to_records_two_start_gt_end_col_large_overage(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=12,
            start_col=50,
            end_col=49
        )
        assert records == [{}, {}, {}]

    def test_to_records_two_start_end_col_equiv(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=12,
            start_col=5,
            end_col=5
        )
        assert records == [{}, {}, {}]


    def test_to_records_two_sorted_start_gt_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            sort=[["a", "desc"]]
        )
        records = view.to_records(
            end_row=12,
            start_col=5,
            end_col=4
        )
        assert records == [{}, {}, {}]

    def test_to_records_two_sorted_start_gt_end_col_large_overage(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            sort=[["a", "desc"]]
        )
        records = view.to_records(
            end_row=12,
            start_col=20,
            end_col=30
        )
        assert records == [{}, {}, {}]

    def test_to_records_two_sorted_start_gt_end_col_overage(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            columns=[],
            group_by=["a"],
            split_by=["b"],
            sort=[["a", "desc"]]
        )
        records = view.to_records(
            end_row=12,
            start_col=1,
            end_col=3
        )
        assert records == [{}, {}, {}]

    def test_to_records_two_sorted_start_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            sort=[["a", "desc"]]
        )
        records = view.to_records(
            start_col=1,
            end_col=2
        )
        assert records == [
            {'2|b': 2, '__ROW_PATH__': []},
            {'2|b': None, '__ROW_PATH__': [3]},
            {'2|b': 2, '__ROW_PATH__': [1]}
        ]

    def test_to_records_two_sorted_start_end_col_equiv(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            sort=[["a", "desc"]]
        )
        records = view.to_records(
            end_row=12,
            start_col=5,
            end_col=5
        )
        assert records == [{}, {}, {}]

    def test_to_records_start_col_end_col(self):
        data = [{"a": 1, "b": 2, "c": 3}, {"a": 3, "b": 4, "c": 5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1,
            end_col=2
        )
        # start_col and end_col access columns at that index - dict key order not guaranteed in python2
        assert records == [{"b": 2}, {"b": 4}]

    def test_to_records_start_col_end_col_equiv(self):
        data = [{"a": 1, "b": 2, "c": 3}, {"a": 3, "b": 4, "c": 5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1,
            end_col=1
        )
        assert records == [{}, {}]

    def test_to_records_floor_start_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1.5
        )
        assert records == [{"b": 2}, {"b": 4}]

    def test_to_records_ceil_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            end_col=1
        )
        assert records == [{"a": 1}, {"a": 3}]

    def test_to_records_two_ceil_end_col(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        records = view.to_records(
            end_row=12,
            end_col=4.5
        )
        assert records == [
            {'2|a': 1, '2|b': 2, '4|a': 3, '4|b': 4, '__ROW_PATH__': []},
            {'2|a': 1, '2|b': 2, '4|a': None, '4|b': None, '__ROW_PATH__': [1]},
            {'2|a': None, '2|b': None, '4|a': 3, '4|b': 4, '__ROW_PATH__': [3]}
        ]

    def test_to_records_floor_start_col_ceil_end_col(self):
        data = [{"a": 1, "b": 2, "c": 3}, {"a": 3, "b": 4, "c": 5}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(
            start_col=1.5,
            end_col=1.5
        )
        # start_col and end_col access columns at that index - dict key order not guaranteed in python2
        assert records == [{"b": 2}, {"b": 4}]

    def test_to_dict_start_col_end_col(self):
        data = [{"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 3, "b": 4, "c": 5, "d": 6}]
        tbl = Table(data)
        view = tbl.view()
        d = view.to_dict(
            start_col=1,
            end_col=3
        )
        assert d == {
            "b": [2, 4],
            "c": [3, 5]
        }

    # to csv

    def test_to_csv_symmetric(self):
        csv = "a,b\n1,2\n3,4"
        df = pd.read_csv(StringIO(csv))
        tbl = Table(df)
        view = tbl.view()
        assert view.to_csv() == '"index","a","b"\n0,1,2\n1,3,4\n'

    def test_to_csv_int(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n1,2\n3,4\n'

    def test_to_csv_float(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n1.5,2.5\n3.5,4.5\n'

    def test_to_csv_date(self):
        today = date.today()
        dt_str = today.strftime("%Y-%m-%d")
        data = [{"a": today, "b": 2}, {"a": today, "b": 4}]
        tbl = Table(data)
        assert tbl.schema()["a"] == date
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n{},2\n{},4\n'.format(dt_str, dt_str)

    def test_to_csv_datetime(self):
        dt = datetime(2019, 3, 15, 20, 30, 59, 6000)
        dt_str = dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        data = [{"a": dt, "b": 2}, {"a": dt, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n{},2\n{},4\n'.format(dt_str, dt_str)

    def test_to_csv_bool(self):
        data = [{"a": True, "b": False}, {"a": True, "b": False}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\ntrue,false\ntrue,false\n'

    def test_to_csv_string(self):
        data = [{"a": "string1", "b": "string2"}, {"a": "string3", "b": "string4"}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n"string1","string2"\n"string3","string4"\n'

    def test_to_csv_none(self):
        data = [{"a": None, "b": None}, {"a": None, "b": None}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv() == '"a","b"\n,\n,\n'

    def test_to_csv_custom_rows(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv(start_row=1) == '"a","b"\n3,4\n'

    def test_to_csv_custom_cols(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv(start_col=1) == '"b"\n2\n4\n'

    def test_to_csv_custom_rows_cols(self):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_csv(start_row=1, start_col=1) == '"b"\n4\n'

    def test_to_csv_one(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"]
        )
        assert view.to_csv() == '"a (Group by 1)","a","b"\n,2,4\n1,2,4\n'

    def test_to_csv_two(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"]
        )
        assert view.to_csv() == '"a (Group by 1)","2|a","2|b"\n,2,4\n1,2,4\n'

    def test_to_csv_column_only(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"]
        )
        assert view.to_csv() == '"2|a","2|b"\n1,2\n1,2\n'

    def test_to_csv_one_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            columns=[]
        )
        assert view.to_csv() == '"a (Group by 1)"\n\n1\n'

    def test_to_csv_two_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            group_by=["a"],
            split_by=["b"],
            columns=[]
        )
        assert view.to_csv() == '"a (Group by 1)"\n\n1\n'

    def test_to_csv_column_only_no_columns(self):
        data = [{"a": 1, "b": 2}, {"a": 1, "b": 2}]
        tbl = Table(data)
        view = tbl.view(
            split_by=["b"],
            columns=[]
        )

        assert view.to_csv() == ''

    # implicit index

    def test_to_format_implicit_index_records(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records(index=True) == [
            {"__INDEX__": [0], "a": 1.5, "b": 2.5},
            {"__INDEX__": [1], "a": 3.5, "b": 4.5}
        ]

    def test_to_format_implicit_index_dict(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict(index=True) == {
            "__INDEX__": [[0], [1]],
            "a": [1.5, 3.5],
            "b": [2.5, 4.5]
        }

    def test_to_format_implicit_id_records(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_records(id=True) == [
            {"__ID__": [0], "a": 1.5, "b": 2.5},
            {"__ID__": [1], "a": 3.5, "b": 4.5}
        ]

    def test_to_format_implicit_id_dict(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        assert view.to_dict(id=True) == {
            "__ID__": [[0], [1]],
            "a": [1.5, 3.5],
            "b": [2.5, 4.5]
        }

    def test_to_format_implicit_index_two_dict(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(group_by=["a"], split_by=["b"])
        assert view.to_dict(index=True) == {
            '2.5|a': [1.5, 1.5, None],
            '2.5|b': [2.5, 2.5, None],
            '4.5|a': [3.5, None, 3.5],
            '4.5|b': [4.5, None, 4.5],
            '__INDEX__': [[], [], []],  # index needs to be the same length as each column
            '__ROW_PATH__': [[], [1.5], [3.5]]
        }

    def test_to_format_implicit_index_two_dict(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view(group_by=["a"], split_by=["b"])
        assert view.to_dict(id=True) == {
            '2.5|a': [1.5, 1.5, None],
            '2.5|b': [2.5, 2.5, None],
            '4.5|a': [3.5, None, 3.5],
            '4.5|b': [4.5, None, 4.5],
            '__ID__': [[], [1.5], [3.5]],  # index needs to be the same length as each column
            '__ROW_PATH__': [[], [1.5], [3.5]]
        }

    def test_to_format_implicit_index_np(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data)
        view = tbl.view()
        cols = view.to_numpy(index=True)
        assert np.array_equal(cols["__INDEX__"], np.array([[0], [1]]))

    def test_to_format_explicit_index_records(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data, index="a")
        view = tbl.view()
        assert view.to_records(index=True) == [
            {"__INDEX__": [1.5], "a": 1.5, "b": 2.5},
            {"__INDEX__": [3.5], "a": 3.5, "b": 4.5}
        ]

    def test_to_format_explicit_index_dict(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data, index="a")
        view = tbl.view()
        assert view.to_dict(index=True) == {
            "__INDEX__": [[1.5], [3.5]],
            "a": [1.5, 3.5],
            "b": [2.5, 4.5]
        }

    def test_to_format_explicit_index_np(self):
        data = [{"a": 1.5, "b": 2.5}, {"a": 3.5, "b": 4.5}]
        tbl = Table(data, index="a")
        view = tbl.view()
        cols = view.to_numpy(index=True)
        assert np.array_equal(cols["__INDEX__"], np.array([[1.5], [3.5]]))

    def test_to_format_explicit_index_str_records(self):
        data = [{"a": "a", "b": 2.5}, {"a": "b", "b": 4.5}]
        tbl = Table(data, index="a")
        view = tbl.view()
        assert view.to_records(index=True) == [
            {"__INDEX__": ["a"], "a": "a", "b": 2.5},
            {"__INDEX__": ["b"], "a": "b", "b": 4.5}
        ]

    def test_to_format_explicit_index_datetime_records(self):
        data = [{"a": datetime(2019, 7, 11, 9, 0), "b": 2.5}, {"a": datetime(2019, 7, 11, 9, 1), "b": 4.5}]
        tbl = Table(data, index="a")
        view = tbl.view()
        assert view.to_records(index=True) == [
            {"__INDEX__": [datetime(2019, 7, 11, 9, 0)], "a": datetime(2019, 7, 11, 9, 0), "b": 2.5},
            {"__INDEX__": [datetime(2019, 7, 11, 9, 1)], "a": datetime(2019, 7, 11, 9, 1), "b": 4.5}
        ]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import sys
from datetime import date, datetime
from random import randint

from perspective.table import Table


class CustomObjectBlank(object):
    pass

class CustomObjectStore(object):
    def __init__(self, value):
        self._value = value

    def _psp_dtype_(self):
        return "object"

    def __int__(self):
        return int(self._value)

    def __repr__(self):
        return 'test'

class CustomObjectRepr(object):
    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return str(self._value)

class CustomObjectIntPromoteToString(CustomObjectRepr):
    def _psp_dtype_(self):
        return int

class CustomObjectFloatPromoteToString(CustomObjectRepr):
    def _psp_dtype_(self):
        return float

class CustomObjectIntBoth(CustomObjectRepr):
    def _psp_dtype_(self):
        return int

    def _psp_repr_(self):
        return int(self._value) + 1

class CustomObjectFloatBoth(CustomObjectRepr):
    def _psp_dtype_(self):
        return float

    def _psp_repr_(self):
        return float(self._value) + 1.0

class CustomObjectIntConvert(CustomObjectRepr):
    def _psp_dtype_(self):
        return int

    def __int__(self):
        return int(self._value)

    def __repr__(self):
        return 'test'

class CustomObjectFloatConvert(CustomObjectRepr):
    def _psp_dtype_(self):
        return float

    def __float__(self):
        return float(self._value)

    def __repr__(self):
        return 'test'


class CustomObjectIntConvertFromFloat(CustomObjectRepr):
    def _psp_dtype_(self):
        return int

    def __float__(self):
        return float(self._value)

    def __repr__(self):
        return 'test'

class CustomObjectFloatConvertFromInt(CustomObjectRepr):
    def _psp_dtype_(self):
        return float

    def __int__(self):
        return int(self._value)

    def __repr__(self):
        return 'test'

class TestTableObjectsExtract(object):
    def test_table_custom_object(self):
        data = {"a": [CustomObjectBlank()]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

        assert tbl.size() == 1
        assert '<perspective.tests.table.test_table_object.CustomObjectBlank object at 0x' in tbl.view().to_dict()["a"][0]

    def test_table_custom_object_repr(self):
        data = {"a": [CustomObjectRepr(1), CustomObjectRepr(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": ["1", "2"]}

    def test_table_custom_object_repr_update(self):
        data = {"a": [CustomObjectIntBoth(1), CustomObjectIntBoth(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": int}
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": [2, 3]}

        tbl.update([{"a": CustomObjectIntBoth(3)}, {"a": CustomObjectIntBoth(4)}])
        assert tbl.size() == 4
        assert tbl.view().to_dict() == {"a": [2, 3, 4, 5]}

    def test_custom_object_int_promote_to_string(self):
        data = {"a": [CustomObjectIntPromoteToString(1), CustomObjectIntPromoteToString(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": ["1", "2"]}
   
    def test_custom_object_float_promote_to_string(self):
        data = {"a": [CustomObjectFloatPromoteToString(1), CustomObjectFloatPromoteToString(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": ["1", "2"]}
   
    def test_custom_object_int_both(self):
        data = {"a": [CustomObjectIntBoth(1), CustomObjectIntBoth(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": int}

        assert tbl.size() == 2
        # We do value + 1 just to make sure
        assert tbl.view().to_dict() == {"a": [2, 3]}

    def test_custom_object_float_both(self):
        data = {"a": [CustomObjectFloatBoth(1), CustomObjectFloatBoth(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": float}

        assert tbl.size() == 2
        # We do value + 1 just to make sure
        assert tbl.view().to_dict() == {"a": [2.0, 3.0]}

    def test_custom_object_int_convert(self):
        data = {"a": [CustomObjectIntConvert(1), CustomObjectIntConvert(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": int}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": [1, 2]}

    def test_custom_object_float_convert(self):
        data = {"a": [CustomObjectFloatConvert(1), CustomObjectFloatConvert(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": float}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": [1.0, 2.0]}

    def test_custom_object_int_convert_from_float(self):
        data = {"a": [CustomObjectIntConvertFromFloat(1), CustomObjectIntConvertFromFloat(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": int}

        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": [1, 2]}

    def test_custom_object_float_convert_from_int(self):
        data = {"a": [CustomObjectFloatConvertFromInt(1), CustomObjectFloatConvertFromInt(2)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": float}
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {"a": [1.0, 2.0]}
    
class TestTableObjectsStore(object):
    def test_object_passthrough(self):
        t = CustomObjectStore(1)
        t2 = CustomObjectStore(2)
        t3 = CustomObjectStore(3)

        data = {"a": [t, t2, t3]}
        tbl = Table(data)
        assert tbl.schema() == {"a": object}
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {"a": [t, t2, t3]}
    

    def test_object_referencecount(self):
        t = CustomObjectStore(1)
        data = {"a": [t]}
        tbl = Table(data)
        assert tbl.schema() == {"a": object}
        assert tbl.size() == 1
        assert tbl.view().to_dict() == {"a": [t]}

        # Count references
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount, and one for the table
        assert sys.getrefcount(t) == 4

    def test_object_referencecount_update(self):
        t = CustomObjectStore(1)
        data = {"a": [t]}
        tbl = Table(data)
        assert tbl.schema() == {"a": object}
        assert tbl.size() == 1
        assert tbl.view().to_dict() == {"a": [t]}

        # Count references
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount, and one for the table
        assert sys.getrefcount(t) == 4

        count = randint(5, 10)
        for c in range(count):
            tbl.update([data])
            # c+1 new copies, +1 for original
            assert tbl.size() == (c+1) + 1
            # c+1 copies in the table now, +1 for original and +3 for others (`t`, `data`, arg to getrefcount)
            assert sys.getrefcount(t) == (c+1) + 4

    def test_object_referencecount_clear(self):
        t = CustomObjectStore(1)
        data = {"a": [t]}
        tbl = Table(data)
        assert tbl.schema() == {"a": object}
        assert tbl.size() == 1

        # Count references
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount, and one for the table
        assert sys.getrefcount(t) == 4

        tbl.clear()
        assert tbl.size() == 0
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount
        assert sys.getrefcount(t) == 3


    def test_object_referencecount_update_clear(self):
        t = CustomObjectStore(1)
        data = {"a": [t]}
        tbl = Table(data)
        assert tbl.schema() == {"a": object}
        assert tbl.size() == 1
        assert tbl.view().to_dict() == {"a": [t]}

        # Count references
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount, and one for the table
        assert sys.getrefcount(t) == 4

        # do random number of updates
        count = randint(5, 10)
        for _ in range(count):
            tbl.update([data])

        tbl.clear()
        assert tbl.size() == 0
        assert tbl.view().to_dict() == {}
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount
        assert sys.getrefcount(t) == 3

    def test_object_referencecount_update_index(self):
        t = CustomObjectStore(1)
        data = {"a": [0], "b": [t]}
        tbl = Table(data, index="a")
        assert tbl.schema() == {"a": int, "b": object}
        assert tbl.size() == 1
        assert tbl.view().to_dict() == {"a": [0], "b": [t]}

        # Count references
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount, and one for the table
        assert sys.getrefcount(t) == 4

        # do random number of updates
        count = randint(5, 10)
        for _ in range(count):
            tbl.update([data])

        # unchanged
        assert tbl.size() == 1
        assert sys.getrefcount(t) == 4

        tbl.clear()
        assert tbl.size() == 0
        assert tbl.view().to_dict() == {}
        # 1 for `t`, one for `data`, one for argument to sys.getrefcount
        assert sys.getrefcount(t) == 3

    def test_object_referencecount_update_complicatedsequence(self):
        from .object_sequence import run
        run()

    def test_object_referencecount_delete(self):
        t = CustomObjectStore(1)
        t2 = CustomObjectStore(2)
        t_ref_count = 2
        t2_ref_count = 2

        tbl = Table({"a": [1], "b": [t]}, index="a")
        t_ref_count += 1

        assert tbl.schema() == {"a": int, "b": object}
        assert tbl.size() == 1

        # Count references
        # 1 for `t`, 1 for `data`, 1 for argument to sys.getrefcount, and 1 for the table
        print(sys.getrefcount(t), "should be", t_ref_count)
        print(sys.getrefcount(t2), "should be", t2_ref_count)

        tbl.update({"a": [2], "b": [t]})
        t_ref_count += 1
        tbl.update({"a": [3], "b": [t]})
        t_ref_count += 1

        assert sys.getrefcount(t) == t_ref_count

        tbl.remove([1])

        tbl.clear()
        assert tbl.size() == 0
        assert tbl.view().to_dict() == {}
        print(sys.getrefcount(t), "should be", 2)
        assert sys.getrefcount(t) == 2
        print(sys.getrefcount(t2), "should be", 2)
        assert sys.getrefcount(t2) == 2

    def test_object_referencecount_delete(self):
        from .object_sequence import run2
        run2()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from pytest import raises
from perspective import Table, PerspectiveError, PerspectiveCppError


class TestException(object):
    def test_exception_from_core(self):
        tbl = Table({"a": [1, 2, 3]})

        with raises(PerspectiveCppError) as ex:
            # creating view with unknown column should throw
            tbl.view(group_by=["b"])

        assert (
            str(ex.value)
            == "Invalid column 'b' found in View group_by.\n"
        )

    def test_exception_from_core_catch_generic(self):
        tbl = Table({"a": [1, 2, 3]})
        # `PerspectiveCppError` should inherit from `Exception`
        with raises(Exception) as ex:
            tbl.view(group_by=["b"])

        assert (
            str(ex.value)
            == "Invalid column 'b' found in View group_by.\n"
        )

    def test_exception_from_core_correct_types(self):
        tbl = Table({"a": [1, 2, 3]})

        # `PerspectiveError` should be raised from the Python layer
        with raises(PerspectiveError) as ex:
            tbl.view()
            tbl.delete()

        assert (
            str(ex.value)
            == "Cannot delete a Table with active views still linked to it - call delete() on each view, and try again."
        )

        with raises(PerspectiveCppError) as ex:
            tbl.view(group_by=["b"])

        assert (
            str(ex.value)
            == "Invalid column 'b' found in View group_by.\n"
        )



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .table import Table
from .view import View
from .libbinding import PerspectiveCppError

__all__ = ["Table", "View", "PerspectiveCppError"]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import random
import string
from functools import partial

from ..core.exception import PerspectiveError
from ..table import Table
from .manager_internal import _PerspectiveManagerInternal
from .session import PerspectiveSession


def gen_name(size=10, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for x in range(size))


class PerspectiveManager(_PerspectiveManagerInternal):
    """PerspectiveManager is an orchestrator for running Perspective on the
    server side.

    The core functionality resides in `process()`, which receives
    JSON-serialized messages from a client (implemented by `perspective-viewer`
    in the browser), executes the commands in the message, and returns the
    results of those commands back to the `post_callback`. Table instances
    should be passed to the manager using `host_table` - this allows server
    code to call Table APIs natively instead of proxying them through the
    Manager. Because Perspective is designed to be used in a shared context,
    i.e. multiple clients all accessing the same `Table`, PerspectiveManager
    comes with the context of `sessions` - an encapsulation of the actions
    and resources used by a single connection to Perspective, which can be
    cleaned up after the connection is closed.

    - When a client connects, for example through a websocket, a new session
        should be spawned using `new_session()`.
    - When the websocket closes, call `close()` on the session instance to
        clean up associated resources.
    """

    def __init__(self, lock=False):
        """Create a new ``PerspectiveManager`` instance.

        Keyword Args:
            lock (:obj:`bool`): [description]. Defaults to False.
        """
        super(PerspectiveManager, self).__init__(lock=lock)
        self._loop_callback = None

    def lock(self):
        """Block messages that can mutate the state of :obj:`~perspective.Table`
        objects under management.

        All ``PerspectiveManager`` objects exposed over the internet should be
        locked to prevent content from being mutated by clients.
        """
        self._lock = True

    def unlock(self):
        """Unblock messages that can mutate the state of
        :obj:`~perspective.Table` objects under management."""
        self._lock = False

    def host(self, item, name=None):
        """Given a :obj:`~perspective.Table`, place it under management and
        allow operations on it to be passed through the Manager instance.

        Args:
            item (:obj:`~perspective.Table`) : a Table to be managed.

        Keyword Args:
            name (:obj:`str`) : an optional name to allow retrieval through
                ``get_table``. A name will be generated if not provided.
        """
        name = name or gen_name()
        if isinstance(item, Table):
            self.host_table(name, item)
        else:
            raise PerspectiveError("Only `Table()` instances can be hosted.")

    def host_table(self, name, table):
        """Given a reference to a `Table`, manage it and allow operations on it
        to occur through the Manager.

        If a function for `queue_process` is defined (i.e., by
        :obj:`~perspective.PerspectiveTornadoHandler`), bind the function to
        :obj:`~perspective.Table` and have it call the manager's version of
        `queue_process`.
        """
        if self._loop_callback is not None:
            # always bind the callback to the table's state manager
            self._loop_callback(lambda: table._table.get_pool().set_event_loop())
            table._state_manager.queue_process = partial(
                self._loop_callback, table._state_manager.call_process
            )
        self._tables[name] = table
        return name

    def get_table(self, name):
        """Return a table under management by name."""
        return self._tables.get(name, None)

    def get_table_names(self):
        """Return the tables that are hosted with this manager by name."""
        return list(self._tables.keys())

    def new_session(self):
        return PerspectiveSession(self)

    def call_loop(self, f, *args, **kwargs):
        """Calls `f()` on this `PerspectiveManager`'s event loop if it has one,
        or raise a PerspectiveError if there is no loop callback set using
        `set_loop_callback()`.
        """
        if self._loop_callback is None:
            raise PerspectiveError(
                "Event loop not set on this PerspectiveManager - use set_loop_callback() before calling call_loop()."
            )
        return self._loop_callback(f, *args, **kwargs)

    def set_loop_callback(self, loop_callback):
        """Sets this `PerspectiveManager` to run in Async mode, defering
        `update()` application and releasing the GIL for expensive operations.

        Once called, this `PerspectiveManager` and all Perspective objects it
        hosts must only be interacted with from the same thread.

        Args:
            loop_callback: A function which accepts a function reference and
                its args/kwargs, and schedules it to run on the same thread
                on which set_loop_callback()` was originally invoked.
        """
        if self._loop_callback is not None:
            raise PerspectiveError("PerspectiveManager already has a `loop_callback`")
        if not callable(loop_callback):
            raise PerspectiveError("`loop_callback` must be a function")
        self._loop_callback = loop_callback
        for table in self._tables.values():
            loop_callback(lambda: table._table.get_pool().set_event_loop())
            table._state_manager.queue_process = partial(
                loop_callback, table._state_manager.call_process
            )



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


# This test demonstrates the necessity of locks for server
# responses to multiple clients

# This test should be implemented for every new server handler, but
# should otherwise never fail for existing locking implementations.

# To demonstrate failing behavior, modify the tornado handler like so:
# -        yield self._stream_lock.acquire()
# +        # yield self._stream_lock.acquire()
#          try:
#              yield f(*args, **kwargs)
#          except tornado.websocket.WebSocketClosedError:
#              pass
#          finally:
# -            yield self._stream_lock.release()
# +            ...
# +            # yield self._stream_lock.release()


import asyncio
import pytest
import random

import tornado
from datetime import datetime

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveTornadoHandler,
    tornado_websocket as websocket,
)


data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()

APPLICATION = tornado.web.Application(
    [
        (
            r"/websocket",
            PerspectiveTornadoHandler,
            {"manager": MANAGER, "check_origin": True, "chunk_size": 10},
        )
    ]
)


@pytest.fixture
def app():
    return APPLICATION


class TestPerspectiveTornadoHandlerChunked(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self, port):
        """Connect and initialize a websocket client connection to the
        Perspective tornado server.
        """
        client = await websocket("ws://127.0.0.1:{}/websocket".format(port))
        return client

    @pytest.mark.gen_test(run_sync=False)
    async def test_tornado_handler_lock_inflight(
        self, app, http_client, http_port, sentinel
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(http_port)
        table = client.open_table(table_name)
        views = await asyncio.gather(*[table.view() for _ in range(5)])
        outputs = await asyncio.gather(*[view.to_arrow() for view in views])
        expected = await table.schema()

        for output in outputs:
            assert Table(output).schema(as_string=True) == expected



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import os
import pytest
import sys
import numpy as np
import pandas as pd

from datetime import date, datetime
from functools import partial
from types import MethodType


if os.name == "nt":
    BINDING = "libbinding.pyd"
    PSP = "libpsp.dll"
else:
    BINDING = "libbinding.so"
    PSP = "libpsp.so"

# rename libbinding.so and libpsp.so temporarily to ensure that client mode
# works automatically when the C++ build fails.
lib_path = os.path.join(os.path.dirname(__file__), "..", "..", "table")
binding = os.path.join(lib_path, BINDING)
psp = os.path.join(lib_path, PSP)
new_binding = os.path.join(lib_path, "notlibbinding.so")
new_psp = os.path.join(lib_path, "notlibpsp.so")


def mock_post(self, msg, msg_id=None, assert_msg=None):
    """Mock the widget's `post()` method so we can introspect the contents."""
    assert msg == assert_msg


def unload():
    to_pop = []
    for mod in sys.modules:
        if mod.startswith("perspective"):
            to_pop.append(mod)
    for mod in to_pop:
        sys.modules.pop(mod)


@pytest.fixture(scope="class")
def rename_libraries():
    print("SETTING UP")
    # unload perspective from sys modules
    unload()

    # rename the binding so it doesnt import
    os.rename(binding, new_binding)
    os.rename(psp, new_psp)
    assert os.path.exists(new_binding)
    assert os.path.exists(new_psp)
    assert not os.path.exists(binding)
    assert not os.path.exists(psp)

    # import
    import perspective

    # defer to test
    yield

    # rename back
    os.rename(new_binding, binding)
    os.rename(new_psp, psp)
    assert os.path.exists(binding)
    assert os.path.exists(psp)
    assert not os.path.exists(new_binding)
    assert not os.path.exists(new_psp)

    # unload from sys.modules
    unload()

    # import again
    import perspective


class TestClient(object):
    def test_widget_client(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": [i for i in range(50)]}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == data

    def test_widget_client_np(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {"a": [i for i in range(50)]}

    def test_widget_client_df(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = pd.DataFrame(
            {
                "a": np.arange(10),
                "b": [True for i in range(10)],
                "c": [str(i) for i in range(10)],
            }
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(10)],
            "a": [i for i in range(10)],
            "b": [True for i in range(10)],
            "c": [str(i) for i in range(10)],
        }

    def test_widget_client_date(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": [date(2020, i, 1) for i in range(1, 13)]}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False

        # `data` is mutated at this point, so check against the expected
        # formatting just to make sure.
        assert widget._data == {"a": ["2020-{:02d}-01".format(i) for i in range(1, 13)]}

    def test_widget_client_np_date(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {
            "a": np.array(
                [date(2020, i, 1) for i in range(1, 13)], dtype="datetime64[D]"
            )
        }
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {"a": ["2020-{:02d}-01".format(i) for i in range(1, 13)]}

    def test_widget_client_np_date_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": np.array([date(2020, i, 1) for i in range(1, 13)], dtype="object")}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {"a": ["2020-{:02d}-01".format(i) for i in range(1, 13)]}

    def test_widget_client_df_date(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = pd.DataFrame(
            {"a": [date(2020, i, 1) for i in range(1, 13)]}, dtype="datetime64[ns]"
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(12)],
            "a": ["2020-{:02d}-01 00:00:00".format(i) for i in range(1, 13)],
        }

    def test_widget_client_df_date_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = pd.DataFrame(
            {"a": [date(2020, i, 1) for i in range(1, 13)]}, dtype="object"
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(12)],
            "a": ["2020-{:02d}-01".format(i) for i in range(1, 13)],
        }

    def test_widget_client_datetime(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": [datetime(2020, i, 1, 12, 30, 45) for i in range(1, 13)]}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-{:02d}-01 12:30:45".format(i) for i in range(1, 13)]
        }

    def test_widget_client_np_datetime(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {
            "a": np.array(
                [datetime(2020, i, 1, 12, 30, 45) for i in range(1, 13)],
                dtype="datetime64",
            )
        }
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-{:02d}-01 12:30:45".format(i) for i in range(1, 13)]
        }

    def test_widget_client_np_datetime_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {
            "a": np.array(
                [datetime(2020, i, 1, 12, 30, 45) for i in range(1, 13)], dtype="object"
            )
        }
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-{:02d}-01 12:30:45".format(i) for i in range(1, 13)]
        }

    def test_widget_client_df_datetime(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = pd.DataFrame(
            {"a": [datetime(2020, i, 1, 12, 30, 45) for i in range(1, 13)]},
            dtype="datetime64[ns]",
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(12)],
            "a": ["2020-{:02d}-01 12:30:45".format(i) for i in range(1, 13)],
        }

    def test_widget_client_df_datetime_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = pd.DataFrame(
            {"a": [datetime(2020, i, 1, 12, 30, 45) for i in range(1, 13)]},
            dtype="object",
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(12)],
            "a": ["2020-{:02d}-01 12:30:45".format(i) for i in range(1, 13)],
        }

    def test_widget_client_np_structured_array(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array([(1, 2), (3, 4)], dtype=[("a", "int64"), ("b", "int64")])
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {"a": [1, 3], "b": [2, 4]}

    def test_widget_client_np_recarray(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array([(1, 2), (3, 4)], dtype=[("a", "int64"), ("b", "int64")]).view(
            np.recarray
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {"a": [1, 3], "b": [2, 4]}

    def test_widget_client_np_structured_array_date(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array(
            [
                (date(2020, 1, 1), date(2020, 2, 1)),
                (date(2020, 3, 1), date(2020, 4, 1)),
            ],
            dtype=[("a", "datetime64[D]"), ("b", "datetime64[D]")],
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-01-01", "2020-03-01"],
            "b": ["2020-02-01", "2020-04-01"],
        }

    def test_widget_client_np_recarray_date(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array(
            [
                (date(2020, 1, 1), date(2020, 2, 1)),
                (date(2020, 3, 1), date(2020, 4, 1)),
            ],
            dtype=[("a", "datetime64[D]"), ("b", "datetime64[D]")],
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-01-01", "2020-03-01"],
            "b": ["2020-02-01", "2020-04-01"],
        }

    def test_widget_client_np_structured_array_date_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array(
            [
                (date(2020, 1, 1), date(2020, 2, 1)),
                (date(2020, 3, 1), date(2020, 4, 1)),
            ],
            dtype=[("a", "object"), ("b", "object")],
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-01-01", "2020-03-01"],
            "b": ["2020-02-01", "2020-04-01"],
        }

    def test_widget_client_np_recarray_date_object(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = np.array(
            [
                (date(2020, 1, 1), date(2020, 2, 1)),
                (date(2020, 3, 1), date(2020, 4, 1)),
            ],
            dtype=[("a", "object"), ("b", "object")],
        )
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": ["2020-01-01", "2020-03-01"],
            "b": ["2020-02-01", "2020-04-01"],
        }

    def test_widget_client_schema(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        widget = perspective.PerspectiveWidget(
            {"a": int, "b": float, "c": bool, "d": date, "e": datetime, "f": str}
        )
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": "integer",
            "b": "float",
            "c": "boolean",
            "d": "date",
            "e": "datetime",
            "f": "string",
        }

    def test_widget_client_update(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        comparison_data = {"a": [i for i in range(50)]}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(
            mock_post, assert_msg={"cmd": "update", "data": comparison_data}
        )
        widget.post = MethodType(mocked_post, widget)
        widget.update(data)
        assert hasattr(widget, "table") is False

    def test_widget_client_replace(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        new_data = {"a": [1]}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(
            mock_post, assert_msg={"cmd": "replace", "data": new_data}
        )
        widget.post = MethodType(mocked_post, widget)
        widget.replace(new_data)
        assert widget._data is new_data

    def test_widget_delete_client(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={"cmd": "delete"})
        widget.delete()
        widget.post = MethodType(mocked_post, widget)

    def test_widget_load_split_by_client(self, rename_libraries):
        import perspective

        assert perspective.is_libpsp() is False
        # behavior should not change for client mode
        arrays = [
            np.array(
                [
                    "bar",
                    "bar",
                    "bar",
                    "bar",
                    "baz",
                    "baz",
                    "baz",
                    "baz",
                    "foo",
                    "foo",
                    "foo",
                    "foo",
                    "qux",
                    "qux",
                    "qux",
                    "qux",
                ]
            ),
            np.array(
                [
                    "one",
                    "one",
                    "two",
                    "two",
                    "one",
                    "one",
                    "two",
                    "two",
                    "one",
                    "one",
                    "two",
                    "two",
                    "one",
                    "one",
                    "two",
                    "two",
                ]
            ),
            np.array(
                [
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                    "X",
                    "Y",
                ]
            ),
        ]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
        df_both = pd.DataFrame(
            np.random.randn(3, 16), index=["A", "B", "C"], columns=index
        )
        widget = perspective.PerspectiveWidget(df_both)
        assert hasattr(widget, "table") is False
        assert widget.columns == ["value"]
        assert widget.split_by == ["first", "second", "third"]
        assert widget.group_by == ["index"]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import os
import shutil
import re

HEADER = """---
id: "perspective-python"
title: "perspective-python API"
---

"""

def fix_param_lists(txt):
    return re.sub("^ *?\n    \\* ", "    * ", txt, flags=re.DOTALL | re.MULTILINE)

def fix_headers(txt):
    txt = re.sub("^# ", "## ", txt, flags=re.DOTALL | re.MULTILINE)
    txt = re.sub("^### Examples?", "##### Examples", txt, flags=re.DOTALL | re.MULTILINE)
    return txt

def fix_returns(txt):
    offset = 0
    for match in re.finditer("^\\* \\*\\*Returns\\*\\*.+?\\* \\*\\*Return type\\*\\*", txt, flags=re.DOTALL | re.MULTILINE):
        group = match.group(0)
        new_group = re.sub("        ", "    ", group)
        txt = txt[:match.start(0) - offset] + new_group + txt[match.end(0) - offset:]
        offset += len(group) - len(new_group)
    return txt

def main():
    cwd = os.path.dirname(os.path.realpath(__file__))
    build_dir = os.path.join(cwd, "build")
    if (os.path.exists(build_dir)):
        print("Cleaning docs build folder '{}'".format(build_dir))
        shutil.rmtree(build_dir)
    os.system("sphinx-build -b markdown {} {}".format(cwd, build_dir))

    output = HEADER
    for filename in ["perspective.table.md", "perspective.core.md"]:
        if filename.endswith(".md"):
            with open(os.path.join(build_dir, filename), "r+") as file:
                txt = file.read()
                txt = fix_param_lists(txt)
                txt = fix_headers(txt)
                txt = fix_returns(txt)
                output += txt

    output_dir = os.path.join(cwd, "..", "..", "..", "docs", "docs", "obj")
    with open(os.path.join(output_dir, "perspective-python.md"), "w+") as file:
        file.write(output)
        file.truncate()


main()


################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import asyncio

from functools import partial
from .dispatch import async_queue, subscribe, unsubscribe
from .view_api import view as make_view


def table(client, data, name, index=None, limit=None):
    """Create a Perspective `Table` by posting a message to a Perspective
    server implementation through `client`, returning a `PerspectiveTableProxy`
    object whose API is entirely async and must be called with `await` or
    in a `yield`-based generator."""
    options = {}

    if index:
        options["index"] = index
    elif limit:
        options["limit"] = limit

    msg = {"cmd": "table", "name": name, "args": [data], "options": options}

    future = asyncio.Future()
    client.post(msg, future)
    return future


class PerspectiveTableProxy(object):
    def __init__(self, client, name):
        """A proxy for a Perspective `Table` object elsewhere, i.e. on a remote
        server accessible through a Websocket.

        All public API methods on this proxy are async, and must be called
        with `await` or a `yield`-based coroutine.

        Args:
            client (:obj:`PerspectiveClient`): A `PerspectiveClient` that is
                set up to send messages to a Perspective server implementation
                elsewhere.

            name (:obj:`str`): a `str` name for the Table. Automatically
                generated if using the `table` function defined above.
        """
        self._client = client
        self._name = name
        self._async_queue = partial(async_queue, self._client, self._name)
        self._subscribe = partial(subscribe, self._client, self._name)
        self._unsubscribe = partial(unsubscribe, self._client, self._name)

    def make_port(self):
        return self._async_queue("make_port", "table_method")

    def remove_port(self):
        return self._async_queue("remove_port", "table_method")

    def get_index(self):
        return self._async_queue("get_index", "table_method")

    def get_limit(self):
        return self._async_queue("get_limit", "table_method")

    def clear(self):
        return self._async_queue("clear", "table_method")

    def replace(self, data):
        return self._async_queue("replace", "table_method", data)

    def size(self):
        return self._async_queue("size", "table_method")

    def schema(self, as_string=False):
        return self._async_queue("schema", "table_method", as_string=as_string)

    def expression_schema(self, expressions, **kwargs):
        return self._async_queue(
            "expression_schema", "table_method", expressions, **kwargs
        )

    def columns(self):
        return self._async_queue("columns", "table_method")

    def is_valid_filter(self, filter):
        return self._async_queue("is_valid_filter", "table_method", filter)

    def on_delete(self, callback):
        return self._subscribe("on_delete", "table_method", callback)

    def remove_delete(self, callback):
        return self._unsubscribe("remove_delete", "table_method", callback)

    def delete(self):
        return self._async_queue("delete", "table_method")

    def view(
        self,
        columns=None,
        group_by=None,
        split_by=None,
        aggregates=None,
        sort=None,
        filter=None,
        expressions=None,
    ):
        return make_view(
            self._client,
            self._name,
            columns,
            group_by,
            split_by,
            aggregates,
            sort,
            filter,
            expressions,
        )

    def update(self, data, port_id=0):
        msg = {
            "name": self._name,
            "cmd": "table_method",
            "method": "update",
            "args": [data, {"port_id": port_id}],
            "subscribe": False,
        }
        return self._client.post(msg)

    def remove(self, pkeys, port_id=0):
        msg = {
            "name": self._name,
            "cmd": "table_method",
            "method": "remove",
            "args": [pkeys, {"port_id": port_id}],
            "subscribe": False,
        }
        return self._client.post(msg)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from logging import critical
import os


def is_libpsp():
    """Was libbinding successfully loaded in this module?"""
    return __is_libpsp__


__is_libpsp__ = True

try:
    # Load all `libbinding` depending modules in one go, otherwise nothing
    # dependent on `libbinding` is exposed.
    from .table import *
    from .manager import *
    from .viewer import *
    from .table.libbinding import (
        init_expression_parser,
    )

    def set_threadpool_size(nthreads):
        """Sets the size of the global Perspective thread pool, up to the
        total number of available cores.  Passing an explicit
        `None` sets this limit to the detected hardware concurrency from the
        environment, which is also the default if this method is never called.
        `set_threadpool_size()` must be called before any other
        `perspective-python` API calls, and cannot be changed after such a call.
        """
        os.environ["OMP_THREAD_LIMIT"] = "0" if nthreads is None else str(nthreads)

    init_expression_parser()
except ImportError:
    __is_libpsp__ = False
    critical(
        "Failed to import C++ bindings for Perspective "
        "probably as it could not be built for your architecture "
        "(check install logs for more details).\n",
        exc_info=True,
    )
    critical("You can still use `PerspectiveWidget` in client mode using JupyterLab.")



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import re
from datetime import date, datetime
from functools import partial
from .libbinding import t_dtype

ALIAS_REGEX = re.compile(r"//(.+)\n")
EXPRESSION_COLUMN_NAME_REGEX = re.compile(r"\"(.*?[^\\])\"")
STRING_LITERAL_REGEX = re.compile(r"'(.*?[^\\])'")
FUNCTION_LITERAL_REGEX = re.compile(
    r"(bucket|match|match_all|search|indexof)\(.*?,\s*(intern\(\'(.+)\'\)).*\)"
)
REPLACE_FN_REGEX = re.compile(
    r"(replace_all|replace)\(.*?,\s*(intern\(\'(.*)\'\)),.*\)"
)
BOOLEAN_LITERAL_REGEX = re.compile(r"([a-zA-Z_]+[a-zA-Z0-9_]*)")


def _extract_type(type, typemap):
    rval = typemap.get(type)

    if rval is None:
        return None

    return rval


def _dtype_to_pythontype(dtype):
    """Returns the native Python type from a Perspective type"""
    mapping = {
        t_dtype.DTYPE_BOOL: bool,
        t_dtype.DTYPE_FLOAT32: float,
        t_dtype.DTYPE_FLOAT64: float,
        t_dtype.DTYPE_UINT8: int,
        t_dtype.DTYPE_UINT16: int,
        t_dtype.DTYPE_UINT32: int,
        t_dtype.DTYPE_UINT64: int,
        t_dtype.DTYPE_INT8: int,
        t_dtype.DTYPE_INT16: int,
        t_dtype.DTYPE_INT32: int,
        t_dtype.DTYPE_INT64: int,
        t_dtype.DTYPE_DATE: date,
        t_dtype.DTYPE_TIME: datetime,
        t_dtype.DTYPE_STR: str,
        t_dtype.DTYPE_OBJECT: object,
    }

    return _extract_type(dtype, mapping)


def _dtype_to_str(dtype):
    """Returns the normalized string representation of a Perspective type,
    compatible with Perspective.js.
    """
    mapping = {
        t_dtype.DTYPE_BOOL: "boolean",
        t_dtype.DTYPE_FLOAT32: "float",
        t_dtype.DTYPE_FLOAT64: "float",
        t_dtype.DTYPE_INT8: "integer",
        t_dtype.DTYPE_INT16: "integer",
        t_dtype.DTYPE_INT32: "integer",
        t_dtype.DTYPE_INT64: "integer",
        t_dtype.DTYPE_DATE: "date",
        t_dtype.DTYPE_TIME: "datetime",
        t_dtype.DTYPE_STR: "string",
        t_dtype.DTYPE_OBJECT: "object",
    }

    return _extract_type(dtype, mapping)


def _str_to_pythontype(typestring):
    """Returns a Python type from the normalized string representation of a
    Perspective type, i.e. from Perspective.js.
    """
    mapping = {
        "boolean": bool,
        "float": float,
        "integer": int,
        "date": date,
        "datetime": datetime,
        "string": str,
        "object": object,
    }

    return _extract_type(typestring, mapping)


def _pythontype_to_str(typestring):
    """Returns the normalized string representation of a Perspective type from
    a Python type object.
    """
    mapping = {
        bool: "boolean",
        float: "float",
        int: "integer",
        date: "date",
        datetime: "datetime",
        str: "string",
        object: "object",
    }

    return _extract_type(typestring, mapping)


def _replace_expression_column_name(
    column_name_map, column_id_map, running_cidx, match_obj
):
    """Replace column names in the expression syntax with a column ID,
    filling the column_name_map and column_id_map."""
    column_name = match_obj.group(1)

    # first, replace escaped single quotes inside the column name, assuming that
    # they are escaping a single apostrophe. Because Python treats backslashes
    # as escapes, users need to write two backslashes in order to properly
    # escape single quotes - i.e. "here\\'s an apostrophe".
    if column_name not in column_name_map:
        column_id = "COLUMN{0}".format(running_cidx[0])
        column_name_map[column_name] = column_id
        column_id_map[column_id] = column_name

    running_cidx[0] += 1

    return column_name_map[column_name]


def _replace_interned_param(match_obj):
    """Replace the intern('unit') in `bucket()` with just the string
    literal, because the unit determines the return type of the column and the
    function would not be able to validate a unit if it was interned."""
    full = match_obj.group(0)
    intern_fn = match_obj.group(2)
    value = match_obj.group(3)
    intern_idx = full.index(intern_fn)

    # from fn(param, intern('string'), param, param...) to
    # fn (param, 'string', ...)
    return "{}'{}'{}".format(
        full[0:intern_idx], value, full[intern_idx + len(intern_fn) :]
    )


def _parse_expression_strings(expressions):
    """Given a list of string expressions, parse out column names and string
    literals using regex and return a list of lists that contain three items
    in each inner list:

    0: the original expression that the user typed
    1: the parsed expression that will be computed
    2: a `dict` of column IDs (`COLUMN0`, `COLUMN1` etc.) to actual column
       names, which will be used by the engine.
    """
    validated_expressions = []
    alias_map = {}

    for expression in expressions:
        if '""' in expression:
            raise ValueError("Cannot reference empty column in expression!")

        column_id_map = {}
        column_name_map = {}

        alias_match = re.match(ALIAS_REGEX, expression)

        # initialize `parsed` here so we keep `expression` unedited as the
        # user typed it into Perspective
        parsed = expression

        if alias_match:
            alias = alias_match.group(1).strip()
        else:
            # Expression itself is the alias
            alias = expression

        # we need to be able to modify the running_cidx inside of every call to
        # replacer_fn - must pass by reference unfortunately
        running_cidx = [0]

        replacer_fn = partial(
            _replace_expression_column_name,
            column_name_map,
            column_id_map,
            running_cidx,
        )

        parsed = re.sub(
            BOOLEAN_LITERAL_REGEX,
            lambda match: "True"
            if match.group(0) == "true"
            else ("False" if match.group(0) == "false" else match.group(0)),
            parsed,
        )

        parsed = re.sub(EXPRESSION_COLUMN_NAME_REGEX, replacer_fn, parsed)
        parsed = re.sub(
            STRING_LITERAL_REGEX,
            lambda match: "intern({0})".format(match.group(0)),
            parsed,
        )

        # remove the `intern()` in bucket and regex functions that take
        # string literal parameters. TODO this logic should be centralized
        # in C++ instead of being duplicated.
        parsed = re.sub(FUNCTION_LITERAL_REGEX, _replace_interned_param, parsed)
        parsed = re.sub(REPLACE_FN_REGEX, _replace_interned_param, parsed)

        validated = [alias, expression, parsed, column_id_map]

        if alias_map.get(alias) is not None:
            idx = alias_map[alias]
            validated_expressions[idx] = validated
        else:
            validated_expressions.append(validated)
            alias_map[alias] = len(validated_expressions) - 1

    return validated_expressions



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import os
import random
import uuid
import pyarrow as pa
import pandas as pd
import numpy as np
from datetime import date, datetime
from pytest import mark
from perspective.table import Table

SOURCE_STREAM_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str.arrow")
SOURCE_FILE_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str.arrow")
PARTIAL_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str_update.arrow")
DICT_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "dict.arrow")
DICT_UPDATE_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "dict_update.arrow")

names = ["a", "b", "c", "d"]


class TestUpdateArrow(object):

    # files

    def test_update_arrow_updates_stream_file(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })

        with open(SOURCE_STREAM_ARROW, mode='rb') as file:  # b is important -> binary
            tbl.update(file.read())
            assert tbl.size() == 4
            assert tbl.schema() == {
                "a": int,
                "b": float,
                "c": str
            }

        with open(SOURCE_FILE_ARROW, mode='rb') as file:
            tbl.update(file.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4] * 2,
                "b": [1.5, 2.5, 3.5, 4.5] * 2,
                "c": ["a", "b", "c", "d"] * 2
            }

    def test_update_arrow_partial_updates_file(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        }, index="a")

        with open(SOURCE_STREAM_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(PARTIAL_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4],
                "b": [100.5, 2.5, 3.5, 400.5],
                "c": ["x", "b", "c", "y"]
            }

    def test_update_arrow_updates_dict_file(self):
        tbl = Table({
            "a": str,
            "b": str
        })

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 5

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "def", None, "abc", None, "update1", "update2"],
                "b": ["klm", "hij", None, "hij", "klm", "update3", None, "update4"]
            }

    @mark.skip
    def test_update_arrow_updates_dict_partial_file(self):
        tbl = None
        v = None

        with open(DICT_ARROW, mode='rb') as src:
            tbl = Table(src.read(), index="a")
            v = tbl.view()
            assert v.num_rows() == 2
            assert v.to_dict() == {
                "a": ["abc", "def"],
                "b": ["klm", "hij"]
            }

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            v.num_rows() == 4
            assert v.to_dict() == {
                "a": ["abc", "def", "update1", "update2"],
                "b": ["klm", "hij", None, "update4"]
            }

    # update with file arrow with more columns than in schema

    def test_update_arrow_updates_more_columns_stream_file(self):
        tbl = Table({
            "a": int,
            "b": float,
        })

        with open(SOURCE_STREAM_ARROW, mode='rb') as file:  # b is important -> binary
            tbl.update(file.read())
            assert tbl.size() == 4
            assert tbl.schema() == {
                "a": int,
                "b": float
            }

        with open(SOURCE_FILE_ARROW, mode='rb') as file:
            tbl.update(file.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4] * 2,
                "b": [1.5, 2.5, 3.5, 4.5] * 2
            }

    def test_update_arrow_partial_updates_more_columns_file(self):
        tbl = Table({
            "a": int,
            "c": str
        }, index="a")

        with open(SOURCE_STREAM_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(PARTIAL_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4],
                "c": ["x", "b", "c", "y"]
            }

    def test_update_arrow_updates_dict_more_columns_file(self):
        tbl = Table({
            "a": str,
        })

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 5

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "def", None, "abc", None, "update1", "update2"]
            }

    @mark.skip
    def test_update_arrow_updates_dict_more_columns_partial_file(self):
        tbl = Table({
            "a": str
        }, index="a")

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "update1", "update2"]
            }

    # update with file arrow with less columns than in schema

    def test_update_arrow_updates_less_columns_stream_file(self):
        tbl = Table({
            "a": int,
            "x": float,
        })

        with open(SOURCE_STREAM_ARROW, mode='rb') as file:  # b is important -> binary
            tbl.update(file.read())
            assert tbl.size() == 4
            assert tbl.schema() == {
                "a": int,
                "x": float
            }

        with open(SOURCE_FILE_ARROW, mode='rb') as file:
            tbl.update(file.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4] * 2,
                "x": [None for i in range(8)]
            }

    def test_update_arrow_partial_updates_less_columns_file(self):
        tbl = Table({
            "a": int,
            "x": str
        }, index="a")

        with open(SOURCE_STREAM_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(PARTIAL_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4],
                "x": [None for i in range(4)]
            }

    def test_update_arrow_updates_dict_less_columns_file(self):
        tbl = Table({
            "a": str,
            "x": str
        })

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 5

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "def", None, "abc", None, "update1", "update2"],
                "x": [None for i in range(8)]
            }

    @mark.skip
    def test_update_arrow_updates_dict_less_columns_partial_file(self):
        tbl = Table({
            "a": str,
            "x": str
        }, index="a")

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "update1", "update2"],
                "x": [None for i in range(4)]
            }

    # update int schema with int

    def test_update_arrow_update_int_schema_with_uint8(self, util):
        array = [random.randint(0, 127) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint8)
        })

        schema = pa.schema({
            "a": pa.uint8()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_uint16(self, util):
        array = [random.randint(0, 32767) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint16)
        })

        schema = pa.schema({
            "a": pa.uint16()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_uint32(self, util):
        array = [random.randint(0, 2000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint32)
        })

        schema = pa.schema({
            "a": pa.uint32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_uint64(self, util):
        array = [random.randint(0, 20000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint64)
        })

        schema = pa.schema({
            "a": pa.uint64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_int8(self, util):
        array = [random.randint(-127, 127) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int8)
        })

        schema = pa.schema({
            "a": pa.int8()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_int16(self, util):
        array = [random.randint(-32767, 32767) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int16)
        })

        schema = pa.schema({
            "a": pa.int16()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_int32(self, util):
        array = [random.randint(-2000000, 2000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int32)
        })

        schema = pa.schema({
            "a": pa.int32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_int_schema_with_int64(self, util):
        array = [random.randint(-20000000, 20000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int64)
        })

        schema = pa.schema({
            "a": pa.int64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == [x * 1.0 for x in array]

    # updating float schema with int

    def test_update_arrow_update_float_schema_with_uint8(self, util):
        array = [random.randint(0, 127) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint8)
        })

        schema = pa.schema({
            "a": pa.uint8()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_uint16(self, util):
        array = [random.randint(0, 32767) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint16)
        })

        schema = pa.schema({
            "a": pa.uint16()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_uint32(self, util):
        array = [random.randint(0, 2000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint32)
        })

        schema = pa.schema({
            "a": pa.uint32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_uint64(self, util):
        array = [random.randint(0, 20000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.uint64)
        })

        schema = pa.schema({
            "a": pa.uint64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_int8(self, util):
        array = [random.randint(-127, 127) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int8)
        })

        schema = pa.schema({
            "a": pa.int8()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_int16(self, util):
        array = [random.randint(-32767, 32767) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int16)
        })

        schema = pa.schema({
            "a": pa.int16()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_int32(self, util):
        array = [random.randint(-2000000, 2000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int32)
        })

        schema = pa.schema({
            "a": pa.int32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_int64(self, util):
        array = [random.randint(-20000000, 20000000) for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.int64)
        })

        schema = pa.schema({
            "a": pa.int64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    # updating int schema with float
    def test_update_arrow_update_int_schema_with_float32(self, util):
        array = [random.randint(-2000000, 2000000) * 0.5 for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.float32)
        })

        schema = pa.schema({
            "a": pa.float32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == [int(x) for x in array]

    def test_update_arrow_update_int_schema_with_float64(self, util):
        array = [random.randint(-20000000, 20000000) * random.random() for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.float64)
        })

        schema = pa.schema({
            "a": pa.float64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": int
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == [int(x) for x in array]

    # updating float schema with float

    def test_update_arrow_update_float_schema_with_float32(self, util):
        array = [random.randint(-2000000, 2000000) * 0.5 for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.float32)
        })

        schema = pa.schema({
            "a": pa.float32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    def test_update_arrow_update_float_schema_with_float64(self, util):
        array = [random.randint(-20000000, 20000000) * random.random() for i in range(100)]
        data = pd.DataFrame({
            "a": np.array(array, dtype=np.float64)
        })

        schema = pa.schema({
            "a": pa.float64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)
        tbl = Table({
            "a": float
        })
        tbl.update(arrow)
        assert tbl.view().to_dict()["a"] == array

    # updating date schema

    def test_update_arrow_update_date_schema_with_date32(self, util):
        array = [date(2019, 2, i) for i in range(1, 11)]
        data = pd.DataFrame({
            "a": array
        })

        schema = pa.schema({
            "a": pa.date32()
        })

        arrow = util.make_arrow_from_pandas(data, schema)

        tbl = Table({
            "a": date
        })

        tbl.update(arrow)

        assert tbl.view().to_dict()["a"] == [datetime(2019, 2, i) for i in range(1, 11)]

    def test_update_arrow_update_date_schema_with_date64(self, util):
        array = [date(2019, 2, i) for i in range(1, 11)]
        data = pd.DataFrame({
            "a": array
        })

        schema = pa.schema({
            "a": pa.date64()
        })

        arrow = util.make_arrow_from_pandas(data, schema)

        tbl = Table({
            "a": date
        })

        tbl.update(arrow)

        assert tbl.view().to_dict()["a"] == [datetime(2019, 2, i) for i in range(1, 11)]

    def test_update_arrow_update_datetime_schema_with_timestamp(self, util):
        data = [
            [datetime(2019, 2, i, 9) for i in range(1, 11)],
            [datetime(2019, 2, i, 10) for i in range(1, 11)],
            [datetime(2019, 2, i, 11) for i in range(1, 11)],
            [datetime(2019, 2, i, 12) for i in range(1, 11)]
        ]

        arrow_data = util.make_arrow(
            names, data, types=[
                pa.timestamp("s"),
                pa.timestamp("ms"),
                pa.timestamp("us"),
                pa.timestamp("ns"),
            ]
        )

        tbl = Table({
            "a": datetime,
            "b": datetime,
            "c": datetime,
            "d": datetime,
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1],
            "c": data[2],
            "d": data[3]
        }

    # streams

    def test_update_arrow_updates_int_stream(self, util):
        data = [list(range(10)) for i in range(4)]
        arrow_data = util.make_arrow(names, data)
        tbl = Table({
            "a": int,
            "b": int,
            "c": int,
            "d": int
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1],
            "c": data[2],
            "d": data[3]
        }

    def test_update_arrow_updates_float_stream(self, util):
        data = [
            [i for i in range(10)],
            [i * 1.5 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a", "b"], data)
        tbl = Table({
            "a": int,
            "b": float,
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1]
        }

    def test_update_arrow_updates_decimal128_stream(self, util):
        data = [
            [i * 1000000000 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.decimal128(10)])
        tbl = Table({
            "a": int
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_bool_stream(self, util):
        data = [
            [True if i % 2 == 0 else False for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data)
        tbl = Table({
            "a": bool
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_date32_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date32()])
        tbl = Table({
            "a": date
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 2, i) for i in range(1, 11)]
        }

    def test_update_arrow_updates_date64_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date64()])
        tbl = Table({
            "a": date
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 2, i) for i in range(1, 11)]
        }

    def test_update_arrow_updates_timestamp_all_formats_stream(self, util):
        data = [
            [datetime(2019, 2, i, 9) for i in range(1, 11)],
            [datetime(2019, 2, i, 10) for i in range(1, 11)],
            [datetime(2019, 2, i, 11) for i in range(1, 11)],
            [datetime(2019, 2, i, 12) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(
            names, data, types=[
                pa.timestamp("s"),
                pa.timestamp("ms"),
                pa.timestamp("us"),
                pa.timestamp("ns"),
            ]
        )
        tbl = Table({
            "a": datetime,
            "b": datetime,
            "c": datetime,
            "d": datetime
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1],
            "c": data[2],
            "d": data[3]
        }

    def test_update_arrow_updates_string_stream(self, util):
        data = [
            [str(i) for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.string()])
        tbl = Table({
            "a": str
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)
        tbl = Table({
            "a": str,
            "b": str
        })
        tbl.update(arrow_data)

        assert tbl.size() == 4
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None],
            "b": ["x", "y", None, "z"]
        }

    def test_update_arrow_partial_updates_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)

        tbl = Table({
            "a": str,
            "b": str
        }, index="a")

        tbl.update(arrow_data)

        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [None, "a", "b"],
            "b": ["z", "x", "y"]
        }

    @mark.skip
    def test_update_arrow_partial_updates_dictionary_stream_duplicates(self, util):
        """If there are duplicate values in the dictionary, primary keys
        may be duplicated if the column is used as an index. Skip this test
        for now - still looking for the best way to fix."""
        data = [
            ([0, 1, 1, None, 2], ["a", "b", "a"]),
            ([0, 1, None, 2, 1], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)

        tbl = Table({
            "a": str,
            "b": str
        }, index="a")

        tbl.update(arrow_data)

        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [None, "a", "b"],
            "b": ["z", "x", "y"]
        }

    def test_update_arrow_partial_updates_more_columns_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)

        tbl = Table({
            "a": str
        }, index="a")

        tbl.update(arrow_data)

        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [None, "a", "b"]
        }

    def test_update_arrow_partial_updates_less_columns_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)
        tbl = Table({
            "a": str,
            "b": str,
            "x": str
        }, index="a")
        tbl.update(arrow_data)

        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [None, "a", "b"],
            "b": ["z", "x", "y"],
            "x": [None, None, None]
        }

    def test_update_arrow_arbitary_order(self, util):
        data = [[1, 2, 3, 4],
                ["a", "b", "c", "d"],
                [1, 2, 3, 4],
                ["a", "b", "c", "d"]]
        update_data = [[5, 6], ["e", "f"], [5, 6], ["e", "f"]]
        arrow = util.make_arrow(["a", "b", "c", "d"], data)
        update_arrow = util.make_arrow(["c", "b", "a", "d"], update_data)
        tbl = Table(arrow)
        assert tbl.schema() == {
            "a": int,
            "b": str,
            "c": int,
            "d": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 6
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6],
            "b": ["a", "b", "c", "d", "e", "f"],
            "c": [1, 2, 3, 4, 5, 6],
            "d": ["a", "b", "c", "d", "e", "f"]
        }

    # append

    def test_update_arrow_updates_append_int_stream(self, util):
        data = [list(range(10)) for i in range(4)]
        arrow_data = util.make_arrow(names, data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1],
            "c": data[2] + data[2],
            "d": data[3] + data[3]
        }

    def test_update_arrow_updates_append_float_stream(self, util):
        data = [
            [i for i in range(10)],
            [i * 1.5 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a", "b"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1]
        }

    def test_update_arrow_updates_append_decimal_stream(self, util):
        data = [
            [i * 1000 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.decimal128(4)])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_bool_stream(self, util):
        data = [
            [True if i % 2 == 0 else False for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_date32_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        out_data = [datetime(2019, 2, i) for i in range(1, 11)]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date32()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": out_data + out_data
        }

    def test_update_arrow_updates_append_date64_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        out_data = [datetime(2019, 2, i) for i in range(1, 11)]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date64()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": out_data + out_data
        }

    def test_update_arrow_updates_append_timestamp_all_formats_stream(self, util):
        data = [
            [datetime(2019, 2, i, 9) for i in range(1, 11)],
            [datetime(2019, 2, i, 10) for i in range(1, 11)],
            [datetime(2019, 2, i, 11) for i in range(1, 11)],
            [datetime(2019, 2, i, 12) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(
            names, data, types=[
                pa.timestamp("s"),
                pa.timestamp("ms"),
                pa.timestamp("us"),
                pa.timestamp("ns"),
            ]
        )
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1],
            "c": data[2] + data[2],
            "d": data[3] + data[3],
        }

    def test_update_arrow_updates_append_string_stream(self, util):
        data = [
            [str(i) for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.string()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)

        assert tbl.size() == 8
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None, "a", "b", "b", None],
            "b": ["x", "y", None, "z", "x", "y", None, "z"]
        }

    def test_update_arrow_updates_append_dictionary_stream_legacy(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data, legacy=True)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)

        assert tbl.size() == 8
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None, "a", "b", "b", None],
            "b": ["x", "y", None, "z", "x", "y", None, "z"]
        }

    # indexed

    def test_update_arrow_partial_indexed(self, util):
        data = [[1, 2, 3, 4], ["a", "b", "c", "d"]]
        update_data = [[2, 4], ["x", "y"]]
        arrow = util.make_arrow(["a", "b"], data)
        update_arrow = util.make_arrow(["a", "b"], update_data)
        tbl = Table(arrow, index="a")
        assert tbl.schema() == {
            "a": int,
            "b": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 4
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4],
            "b": ["a", "x", "c", "y"]
        }

    # update specific columns

    def test_update_arrow_specific_column(self, util):
        data = [[1, 2, 3, 4], ["a", "b", "c", "d"]]
        update_data = [[2, 3, 4]]
        arrow = util.make_arrow(["a", "b"], data)
        update_arrow = util.make_arrow(["a"], update_data)
        tbl = Table(arrow)
        assert tbl.schema() == {
            "a": int,
            "b": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 7
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 2, 3, 4],
            "b": ["a", "b", "c", "d", None, None, None]
        }

    # try to fuzz column order

    def test_update_arrow_column_order_str(self, util):
        # use str so it doesn't get promoted
        data = [["a", "b", "c"] for i in range(10)]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
        names_scrambled = names[::-1]
        arrow = util.make_arrow(names_scrambled, data)
        tbl = Table({name: str for name in names})
        tbl.update(arrow)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            name: data[0] for name in names
        }

    def test_update_arrow_column_order_int(self, util):
        data = [[1, 2, 3] for i in range(10)]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
        names_scrambled = names[::-1]
        arrow = util.make_arrow(names_scrambled, data)
        tbl = Table({name: int for name in names})
        tbl.update(arrow)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            name: data[0] for name in names
        }

    def test_update_arrow_thread_safe_int_index(self, util):
        data = [["a", "b", "c"] for i in range(10)]
        data += [[1, 2, 3]]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "uid"]
        arrow = util.make_arrow(names, data)
        tbl = Table(arrow, index="uid")

        for i in range(100):
            idx = (1, 2, 3)[random.randint(0, 2)]
            update_data = [[str(uuid.uuid4()) + str(random.randint(100, 1000000000))], [idx]]
            update_names = [names[random.randint(0, 9)], "uid"]
            update_arrow = util.make_arrow(update_names, update_data)
            tbl.update(update_arrow)

        assert tbl.size() == 3

    def test_update_arrow_thread_safe_datetime_index(self, util):
        data = [["a", "b", "c"] for i in range(10)]
        data += [[datetime(2020, 1, 15, 12, 17), datetime(2020, 1, 15, 12, 18), datetime(2020, 1, 15, 12, 19)]]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "uid"]
        arrow = util.make_arrow(names, data)
        tbl = Table(arrow, index="uid")

        for i in range(100):
            idx = (datetime(2020, 1, 15, 12, 17), datetime(2020, 1, 15, 12, 18), datetime(2020, 1, 15, 12, 19))[random.randint(0, 2)]
            update_data = [[str(uuid.uuid4()) + str(random.randint(100, 1000000000))], [idx]]
            update_names = [names[random.randint(0, 9)], "uid"]
            update_arrow = util.make_arrow(update_names, update_data)
            tbl.update(update_arrow)

        assert tbl.size() == 3

    def test_update_arrow_thread_safe_str_index(self, util):
        data = [["a", "b", "c"] for i in range(11)]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "uid"]
        arrow = util.make_arrow(names, data)
        tbl = Table(arrow, index="uid")

        for i in range(100):
            idx = ("a", "b", "c")[random.randint(0, 2)]
            update_data = [[str(uuid.uuid4()) + str(random.randint(100, 1000000000))], [idx]]
            update_names = [names[random.randint(0, 9)], "uid"]
            update_arrow = util.make_arrow(update_names, update_data)
            tbl.update(update_arrow)

        assert tbl.size() == 3



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .common import PerspectiveHandlerBase


class PerspectiveStarletteHandler(PerspectiveHandlerBase):
    """PerspectiveStarletteHandler is a drop-in implementation of Perspective.

    The Perspective client and server will automatically keep the Websocket
    alive without timing out.

    Examples:
        >>> MANAGER = PerspectiveManager()
        >>> MANAGER.host_table("data_source_one", Table(
        ...     pd.read_csv("superstore.csv")))
        >>> app = FastAPI()
        >>> async def endpoint(websocket: Websocket):
        ...     handler = PerspectiveStarletteHandler(manager, websocket)
        ...     await handler.run()
        ... app.add_api_websocket_route('/websocket', endpoint)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._websocket = kwargs["websocket"]

    async def run(self) -> None:
        try:
            await self._websocket.accept()
            while True:
                message = await self._websocket.receive()
                self._websocket._raise_on_disconnect(message)
                if "text" in message:
                    await self.on_message(message["text"])
                if "bytes" in message:
                    await self.on_message(message["bytes"])
        finally:
            self.on_close()

    async def write_message(self, message: str, binary: bool = False) -> None:
        if binary:
            await self._websocket.send_bytes(message)
        else:
            await self._websocket.send_text(message)

    # Use common docstring
    __init__.__doc__ = PerspectiveHandlerBase.__init__.__doc__



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from pytest import raises
from perspective import PerspectiveError, PerspectiveViewer,\
                        PerspectiveWidget, Sort


class TestSort(object):

    def test_sort_widget_load(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data, sort=[["a", Sort.DESC]])
        assert widget.sort == [["a", "desc"]]

    def test_sort_widget_setattr(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        widget.sort = [["a", Sort.ASC_ABS]]
        assert widget.sort == [["a", "asc abs"]]

    def test_sort_widget_load_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        with raises(PerspectiveError):
            PerspectiveWidget(data, sort=[["a", "?"]])

    def test_sort_widget_setattr_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        with raises(PerspectiveError):
            widget.sort = [["a", "?"]]

    def test_sort_widget_init_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        for sort in Sort:
            widget = PerspectiveWidget(data, sort=[["a", sort]])
            assert widget.sort == [["a", sort.value]]

    def test_sort_widget_set_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        for sort in Sort:
            widget.sort = [["a", sort]]
            assert widget.sort == [["a", sort.value]]

    def test_sort_viewer_load(self):
        viewer = PerspectiveViewer(sort=[["a", Sort.COL_ASC_ABS]])
        assert viewer.sort == [["a", "col asc abs"]]

    def test_sort_viewer_setattr(self):
        viewer = PerspectiveViewer()
        viewer.sort = [["a", Sort.COL_DESC_ABS]]
        assert viewer.sort == [["a", "col desc abs"]]

    def test_sort_viewer_init_all(self):
        for sort in Sort:
            viewer = PerspectiveViewer(sort=[["a", sort]])
            assert viewer.sort == [["a", sort.value]]

    def test_sort_viewer_set_all(self):
        viewer = PerspectiveViewer()
        for sort in Sort:
            viewer.sort = [["a", sort]]
            assert viewer.sort == [["a", sort.value]]
